"""Conversion examples python side."""
import numpy as np

import example_carma as carma

sample = np.asarray(
    np.random.random(size=(10, 2)),
    dtype=np.float64,
    order='F'
)

print(carma.manual_example(sample))
print(carma.automatic_example(sample))


sample2 = np.asarray(
    np.random.random(size=(10, 2)),
    dtype=np.float64,
    order='F'
)

example_class = carma.ExampleClass(sample, sample2)
arr = example_class.member_func()
print(arr)

print('\n\n OLS Example \n\n')
y = np.linspace(1, 100, num=100) + np.random.normal(0, 0.5, 100)
X = np.hstack((
    np.ones(100)[:, None],
    np.arange(1, 101)[:, None]
))
coeff, std_err = carma.ols(y, X)
print('coefficients: ', coeff.flatten())
print('std errors: ', std_err.flatten())



import unittest
import pandas as pd
import numpy as np
import time

from banditpam import KMedoids
from utils import bpam_agrees_pam
from constants import (
    NUM_MEDIUM_CASES,
    SMALL_K_SCHEDULE,
    N_SMALL_K,
    MEDIUM_SAMPLE_SIZE,
    LARGE_SAMPLE_SIZE,
    MEDIUM_SIZE_SCHEDULE,
    NUM_MEDIUM_SIZES,
    MNIST_SIZE_MULTIPLIERS,
    SCRNA_SIZE_MULTIPLIERS,
    PROPORTION_PASSING,
    SCALING_EXPONENT,
)


class LargerTests(unittest.TestCase):
    mnist_70k = pd.read_csv("data/MNIST_70k.csv", sep=" ", header=None)
    scrna = pd.read_csv("data/scrna_reformat.csv.gz", header=None)

    def test_medium_mnist(self):
        """
        Generate NUM_MEDIUM_CASES random subsets of MNIST of size
        MEDIUM_SAMPLE_SIZE and verify BanditPAM agrees with PAM.
        """
        num_succeed = 0
        for i in range(NUM_MEDIUM_CASES):
            data = self.mnist_70k.sample(n=MEDIUM_SAMPLE_SIZE).to_numpy()
            num_succeed += bpam_agrees_pam(
                k=SMALL_K_SCHEDULE[i % N_SMALL_K],
                data=data,
                loss="L2",
                test_build=True,
                assert_immediately=False,
            )
        self.assertTrue(
            num_succeed >= PROPORTION_PASSING * NUM_MEDIUM_CASES
        )  # avoids stochasticity issues

    def test_various_medium_mnist(self):
        """
        Generate NUM_MEDIUM_CASES random subsets of MNIST of various sizes and
        verify BanditPAM agrees with PAM.

        Since PAM is very slow, we can only do this for fairly small sizes in
        MEDIUM_SIZE_SCHEDULE
        """
        num_succeed = 0
        for i in range(NUM_MEDIUM_CASES):
            size = MEDIUM_SIZE_SCHEDULE[i % NUM_MEDIUM_SIZES]
            data = self.mnist_70k.sample(n=size)
            num_succeed += bpam_agrees_pam(
                k=SMALL_K_SCHEDULE[i % N_SMALL_K],
                data=data,
                loss="L2",
                test_build=True,
                assert_immediately=False,
            )
        self.assertTrue(
            num_succeed >= PROPORTION_PASSING * NUM_MEDIUM_CASES
        )  # avoids stochasticity issues

    def test_time_cases_mnist(self):
        """
        Verify that BanditPAM scales as O(nlogn) on the MNIST dataset.
        """
        MNIST_10k = self.mnist_70k.head(LARGE_SAMPLE_SIZE).to_numpy()
        kmed = KMedoids(n_medoids=5, algorithm="BanditPAM")
        kmed.seed = 0
        start = time.time()
        kmed.fit(MNIST_10k, "L2")
        base_runtime = time.time() - start

        for size_multiplier in MNIST_SIZE_MULTIPLIERS:
            size = size_multiplier * LARGE_SAMPLE_SIZE
            MNIST_test = self.mnist_70k.head(n=size).to_numpy()
            start = time.time()
            kmed.fit(MNIST_test, "L2")
            runtime = time.time() - start
            # TODO(@motiwari): Timing test will not work
            # need to compute it over number of steps
            self.assertTrue(
                runtime < (size_multiplier**SCALING_EXPONENT) * base_runtime
            )

    def test_medium_scrna(self):
        """
        Generate NUM_MEDIUM_CASES random subsets of MNIST of size
        MEDIUM_SAMPLE_SIZE and verify BanditPAM agrees with PAM.
        """
        count = 0
        for i in range(NUM_MEDIUM_CASES):
            data = self.scrna.sample(n=MEDIUM_SAMPLE_SIZE).to_numpy()
            count += bpam_agrees_pam(
                k=SMALL_K_SCHEDULE[i % N_SMALL_K],
                data=data,
                loss="L1",
                test_build=False,
                assert_immediately=False,
            )
        self.assertTrue(count >= PROPORTION_PASSING * NUM_MEDIUM_CASES)

    def test_various_medium_scrna(self):
        """
        Generate NUM_MEDIUM_CASES random subsets of scRNA of various sizes and
        verify BanditPAM agrees with PAM.

        Since PAM is very slow, we can only do this for fairly small sizes in
        MEDIUM_SIZE_SCHEDULE
        """
        count = 0
        for i in range(NUM_MEDIUM_CASES):
            size = MEDIUM_SIZE_SCHEDULE[i % NUM_MEDIUM_SIZES]
            data = self.scrna.sample(n=size)
            count += bpam_agrees_pam(
                k=SMALL_K_SCHEDULE[i % N_SMALL_K],
                data=data,
                loss="L2",
                test_build=False,
                assert_immediately=False,
            )
        self.assertTrue(count >= PROPORTION_PASSING * NUM_MEDIUM_CASES)

    def test_time_cases_scrna(self):
        """
        Verify that BanditPAM scales as O(nlogn) on the MNIST dataset.
        """
        SCRNA_10k = self.scrna.head(LARGE_SAMPLE_SIZE).to_numpy()
        kmed = KMedoids(n_medoids=5, algorithm="BanditPAM")
        kmed.seed = 0
        start = time.time()
        kmed.fit(SCRNA_10k, "L1")
        base_runtime = time.time() - start

        for size_multiplier in SCRNA_SIZE_MULTIPLIERS:
            size = size_multiplier * LARGE_SAMPLE_SIZE
            scrna_test = self.scrna.head(n=size).to_numpy()
            start = time.time()
            kmed.fit(scrna_test, "L1")
            runtime = time.time() - start
            # TODO(@motiwari): Timing test will not work
            # need to compute it over number of steps
            self.assertTrue(
                runtime < (size_multiplier**SCALING_EXPONENT) * base_runtime
            )


if __name__ == "__main__":
    np.random.seed(0)
    unittest.main()



import unittest

from banditpam import KMedoids


class InitializationTests(unittest.TestCase):
    def test_default_initialization(self):
        kmed = KMedoids()
        assert kmed.n_medoids == 5, "n_medoids does not match"
        assert kmed.algorithm == "BanditPAM", "algorithm does not match"
        assert kmed.max_iter == 100, "max_iter does not match"
        assert kmed.build_confidence == 10, "build_confidence does not match"
        assert kmed.swap_confidence == 5, "swap_confidence does not match"
        assert kmed.use_cache, "use_cache does not match"
        assert kmed.use_perm, "use_perm does not match"
        assert kmed.cache_width == 1000, "cache_width does not match"
        assert kmed.parallelize, "parallelize does not match"

    def test_initialization_1(self):
        """
        Test that we can initialize a KMedoids object with given parameters.
        """
        kmed = KMedoids(
            n_medoids=5,
            algorithm="BanditPAM",
            max_iter=100,
            build_confidence=9,
            swap_confidence=11,
            use_cache=True,
            use_perm=True,
            cache_width=10,
            parallelize=True,
        )
        assert kmed.n_medoids == 5, "n_medoids does not match"
        assert kmed.algorithm == "BanditPAM", "algorithm does not match"
        assert kmed.max_iter == 100, "max_iter does not match"
        assert kmed.build_confidence == 9, "build_confidence does not match"
        assert kmed.swap_confidence == 11, "swap_confidence does not match"
        assert kmed.use_cache, "use_cache does not match"
        assert kmed.use_perm, "use_perm does not match"
        assert kmed.cache_width == 10, "cache_width does not match"
        assert kmed.parallelize, "parallelize does not match"

    def test_initialization_2(self):
        kmed = KMedoids(
            n_medoids=3,
            algorithm="FastPAM1",
            max_iter=1,
            build_confidence=11,
            swap_confidence=2,
            use_cache=False,
            use_perm=True,
            cache_width=60,
            parallelize=False,
        )
        assert kmed.n_medoids == 3, "n_medoids does not match"
        assert kmed.algorithm == "FastPAM1", "algorithm does not match"
        assert kmed.max_iter == 1, "max_iter does not match"
        assert kmed.build_confidence == 11, "build_confidence does not match"
        assert kmed.swap_confidence == 2, "swap_confidence does not match"
        assert not kmed.use_cache, "use_cache does not match"
        assert kmed.use_perm, "use_perm does not match"
        assert kmed.cache_width == 60, "cache_width does not match"
        assert not kmed.parallelize, "parallelize does not match"

    def test_initialization_3(self):
        kmed = KMedoids(
            n_medoids=100,
            algorithm="PAM",
            max_iter=10000,
            build_confidence=11000,
            swap_confidence=2029,
            use_cache=True,
            use_perm=False,
            cache_width=60000,
            parallelize=True,
        )
        assert kmed.n_medoids == 100, "n_medoids does not match"
        assert kmed.algorithm == "PAM", "algorithm does not match"
        assert kmed.max_iter == 10000, "max_iter does not match"
        assert (
            kmed.build_confidence == 11000
        ), "build_confidence does not match"
        assert kmed.swap_confidence == 2029, "swap_confidence does not match"
        assert kmed.use_cache, "use_cache does not match"
        assert not kmed.use_perm, "use_perm does not match"
        assert kmed.cache_width == 60000, "cache_width does not match"
        assert kmed.parallelize, "parallelize does not match"


if __name__ == "__main__":
    unittest.main()



import sys
import numpy as np
import test_carma as tc

sample = np.random.normal(0, 1, (1000, 2))
fsample = np.asarray(sample, order='F')
ref_sample = sample.sum()

print('------ C-order -- Borrow ------')
print('ref count sample -- pre: ', sys.getrefcount(sample))
print('ref count fsample -- pre: ', sys.getrefcount(fsample))
arr = tc.debug_arr_to_mat(sample, 0, 0)
print('ref count sample -- post: ', sys.getrefcount(sample))
print('ref count fsample -- post: ', sys.getrefcount(fsample))
assert np.isclose(ref_sample, arr.sum())

print('------ C-order -- Copy ------')
print('ref count sample -- pre: ', sys.getrefcount(sample))
print('ref count fsample -- pre: ', sys.getrefcount(fsample))
arr = tc.debug_arr_to_mat(sample, 1, 0)
print('ref count sample -- post: ', sys.getrefcount(sample))
print('ref count fsample -- post: ', sys.getrefcount(fsample))
assert np.isclose(ref_sample, arr.sum())

print('------ C-order -- Steal ------')
print('ref count sample -- pre: ', sys.getrefcount(sample))
print('ref count fsample -- pre: ', sys.getrefcount(fsample))
arr = tc.debug_arr_to_mat(sample, -1, 0)
print('ref count sample -- post: ', sys.getrefcount(sample))
print('ref count fsample -- post: ', sys.getrefcount(fsample))
assert np.isclose(ref_sample, arr.sum())

print('------ F-order -- Borrow ------')
print('ref count sample -- pre: ', sys.getrefcount(sample))
print('ref count fsample -- pre: ', sys.getrefcount(fsample))
arr = tc.debug_arr_to_mat(fsample, 0, 0)
print('ref count sample -- post: ', sys.getrefcount(sample))
print('ref count fsample -- post: ', sys.getrefcount(fsample))
assert np.isclose(ref_sample, arr.sum())

print('------ F-order -- Copy ------')
print('ref count sample -- pre: ', sys.getrefcount(sample))
print('ref count fsample -- pre: ', sys.getrefcount(fsample))
arr = tc.debug_arr_to_mat(fsample, 1, 0)
print('ref count sample -- post: ', sys.getrefcount(sample))
print('ref count fsample -- post: ', sys.getrefcount(fsample))
assert np.isclose(ref_sample, arr.sum())

print('------ F-order -- Steal ------')
print('ref count sample -- pre: ', sys.getrefcount(sample))
print('ref count fsample -- pre: ', sys.getrefcount(fsample))
arr = tc.debug_arr_to_mat(fsample, -1, 0)
print('ref count sample -- post: ', sys.getrefcount(sample))
print('ref count fsample -- post: ', sys.getrefcount(fsample))
assert np.isclose(ref_sample, arr.sum())



"""Test armadillo matrix to numpy array type caster."""
import numpy as np

import test_carma as carma


def test_type_caster_in_mat():
    """Test type caster input handling of matrix."""
    sample = np.random.normal(size=(25, 2))
    npsum = sample.sum()
    accu = carma.tc_in_mat(sample)
    assert np.isclose(accu, npsum)

    sample = np.random.normal(size=(2, 25))
    npsum = sample.sum()
    accu = carma.tc_in_mat(sample)
    assert np.isclose(accu, npsum)


def test_type_caster_in_row():
    """Test type caster input handling of row."""
    sample = np.random.normal(size=(25))
    npsum = sample.sum()
    accu = carma.tc_in_row(sample)
    assert np.isclose(accu, npsum)


def test_type_caster_in_row_2d():
    """Test type caster input handling of 2d row."""
    sample = np.random.normal(size=(1, 25))
    npsum = sample.sum()
    accu = carma.tc_in_row(sample)
    assert np.isclose(accu, npsum)


def test_type_caster_in_col():
    """Test type caster input handling of column."""
    sample = np.random.normal(size=(25))
    npsum = sample.sum()
    accu = carma.tc_in_col(sample)
    assert np.isclose(accu, npsum)


def test_type_caster_in_col_2d():
    """Test type caster input handling of 2d column."""
    sample = np.random.normal(size=(25, 1))
    npsum = sample.sum()
    accu = carma.tc_in_col(sample)
    assert np.isclose(accu, npsum)


def test_type_caster_in_cube():
    """Test type caster input handling of cube."""
    sample = np.random.normal(size=(25, 2, 3))
    npsum = sample.sum()
    accu = carma.tc_in_cube(sample)
    assert np.isclose(accu, npsum)


def test_type_caster_out_mat():
    """Test type caster output handling of matrix."""
    sample = np.random.normal(size=(25, 2))
    mat = carma.tc_out_mat(sample)
    assert np.allclose(mat, sample + 1)

def test_type_caster_out_mat_const():
    """Test type caster output handling of matrix."""
    sample = np.random.normal(size=(25, 2))
    mat = carma.tc_out_mat_const(sample)
    assert np.allclose(mat, sample + 1)


def test_type_caster_out_mat_rvalue():
    """Test type caster output handling of matrix rvalue."""
    sample = np.random.normal(size=(25, 2))
    mat = carma.tc_out_mat_rvalue(sample)
    assert np.allclose(mat, sample + 1)


def test_type_caster_out_row():
    """Test type caster input handling of matrix."""
    sample = np.random.normal(size=(25))
    mat = carma.tc_out_row(sample)
    assert np.allclose(mat, 1 + sample)


def test_type_caster_out_row_2d():
    """Test type caster input handling of matrix."""
    sample = np.asarray(np.random.normal(size=(1, 25)), order='F')
    mat = carma.tc_out_row(sample)
    assert np.allclose(mat, 1 + sample)


def test_type_caster_out_row_rvalue():
    """Test type caster input handling of matrix."""
    sample = np.random.normal(size=(25))
    mat = carma.tc_out_row_rvalue(sample)
    assert np.allclose(mat, 1 + sample)


def test_type_caster_out_col():
    """Test type caster input handling of matrix."""
    sample = np.asarray(np.random.normal(size=(25)), order='F')
    mat = carma.tc_out_col(sample)
    assert np.allclose(mat.flatten(), 1 + sample)


def test_type_caster_out_col_2d():
    """Test type caster input handling of matrix."""
    sample = np.asarray(np.random.normal(size=(25, 1)), order='F')
    mat = carma.tc_out_col(sample)
    assert np.allclose(mat, 1 + sample)


def test_type_caster_out_col_rvalue():
    """Test type caster input handling of matrix."""
    sample = np.asarray(np.random.normal(size=(25)), order='F')
    mat = carma.tc_out_col_rvalue(sample)
    assert np.allclose(mat.flatten(), 1 + sample)


def test_type_caster_out_cube():
    """Test type caster input handling of matrix."""
    sample = np.asarray(np.random.normal(size=(25, 3, 2)), order='F')
    mat = carma.tc_out_cube(sample)
    assert np.allclose(mat, 1 + sample)


def test_type_caster_out_cube_rvalue():
    """Test type caster input handling of matrix."""
    sample = np.asarray(np.random.normal(size=(25, 3, 2)), order='F')
    mat = carma.tc_out_cube_rvalue(sample)
    assert np.allclose(mat, 1 + sample)



"""Test nparray.h."""
import numpy as np

import integration_test_carma as carma


def test_is_f_contiguous():
    """Test is_f_contiguous."""
    m = 'F order array should be F contiguous'
    sample = np.ones((10, 2), dtype=np.float64, order='F')
    assert carma.is_f_contiguous(sample) is True, m

    m = 'C order array should not be F contiguous'
    sample = np.ones((10, 2), dtype=np.float64, order='C')
    assert carma.is_f_contiguous(sample) is False, m


def test_mat_roundtrip():
    """Test mat_roundtrip."""
    og_sample = np.asarray(
        np.random.normal(size=(50, 2)), dtype=np.float64, order='F'
    )
    sample = og_sample.copy()
    out = carma.mat_roundtrip(sample)
    assert np.allclose(og_sample, out)



"""Test numpy array to matrix conversion function."""
import numpy as np
import pytest

import test_carma as carma

test_flags = {
    1: 'Number of elements between array and matrix are not the same',
    2: 'Number of rows between array and matrix are not the same',
    3: 'Number of columns between array and matrix are not the same',
    4: 'Sum of elements between array and matrix is not aproximately equal',
    5: 'Pointer to memory is not as expected',
}


def test_arr_to_mat_double():
    """Test arr_to_mat."""
    sample = np.asarray(
        np.random.normal(size=(50, 2)), dtype=np.float64, order='F'
    )
    flag = carma.arr_to_mat_double(sample, False)
    assert flag == 0, test_flags[flag]


def test_arr_to_mat_double_large():
    """Test arr_to_mat."""
    sample = np.asarray(
        np.random.normal(size=(1000, 1000)), dtype=np.float64, order='F'
    )
    flag = carma.arr_to_mat_double(sample, False)
    assert flag == 0, test_flags[flag]


def test_arr_to_mat_double_small():
    """Test arr_to_mat."""
    sample = np.asarray(
        np.random.normal(size=(3, 3)), dtype=np.float64, order='F'
    )
    flag = carma.arr_to_mat_double(sample, False)
    assert flag == 0, test_flags[flag]


def test_arr_to_mat_long():
    """Test arr_to_mat."""
    sample = np.asarray(
        np.random.normal(size=(50, 2)), dtype=np.int64, order='F'
    )
    flag = carma.arr_to_mat_long(sample, False)
    assert flag == 0, test_flags[flag]


def test_arr_to_mat_double_c_contiguous():
    """Test arr_to_mat."""
    sample = np.asarray(np.random.normal(size=(50, 2)), dtype=np.float64)
    flag = carma.arr_to_mat_double(sample, False)
    assert flag == 5, test_flags[flag]

def test_arr_to_mat_double_c_contiguous_large():
    """Test arr_to_mat."""
    sample = np.asarray(np.random.normal(size=(1000, 1000)), dtype=np.float64)
    flag = carma.arr_to_mat_double(sample, False)
    assert flag == 5, test_flags[flag]


def test_arr_to_mat_double_c_contiguous_small():
    """Test arr_to_mat."""
    sample = np.asarray(np.random.normal(size=(3, 3)), dtype=np.float64)
    flag = carma.arr_to_mat_double(sample, False)
    assert flag == 5, test_flags[flag]


def test_arr_to_mat_long_c_contiguous():
    """Test arr_to_mat."""
    sample = np.asarray(np.random.normal(size=(50, 2)), dtype=np.int64)
    flag = carma.arr_to_mat_long(sample, False)
    assert flag == 5, test_flags[flag]


def test_arr_to_mat_double_copy():
    """Test arr_to_mat."""
    sample = np.asarray(
        np.random.normal(size=(50, 2)), dtype=np.float64, order='F'
    )
    flag = carma.arr_to_mat_double_copy(sample)
    assert flag == 0, test_flags[flag]


def test_arr_to_mat_double_copy_c_contiguous():
    """Test arr_to_mat."""
    sample = np.asarray(
        np.random.normal(size=(50, 2)), dtype=np.float64, order='C'
    )
    flag = carma.arr_to_mat_double_copy(sample)
    assert flag == 0, test_flags[flag]

def test_arr_to_mat_writeable():
    """Test arr_to_mat."""
    sample = np.asarray(
        np.random.normal(size=(50, 2)), dtype=np.float64, order='C'
    )
    sample.setflags(write=0)
    with pytest.raises(RuntimeError):
        flag = carma.arr_to_mat_double(sample, False)

def test_arr_to_mat_writeable_copy():
    """Test arr_to_mat."""
    sample = np.asarray(
        np.random.normal(size=(50, 2)), dtype=np.float64, order='C'
    )
    sample.setflags(write=0)
    flag = carma.arr_to_mat_double(sample, True)
    assert flag == 5, test_flags[flag]

# #############################################################################
#                                   N-DIM 1                                   #
# #############################################################################
def test_arr_to_mat_1d():
    """Test arr_to_mat."""
    sample = np.asarray(
        np.random.normal(size=(100)), dtype=np.float64, order='F'
    )
    flag = carma.arr_to_mat_1d(sample, False)
    assert flag == 0, test_flags[flag]


def test_arr_to_mat_1d_small():
    """Test arr_to_mat."""
    sample = np.asarray(
        np.random.normal(size=5), dtype=np.float64, order='F'
    )
    flag = carma.arr_to_mat_1d(sample, False)
    assert flag == 0, test_flags[flag]


def test_arr_to_mat_1d_copy():
    """Test arr_to_mat."""
    sample = np.asarray(
        np.random.normal(size=(100)), dtype=np.float64, order='F'
    )
    flag = carma.arr_to_mat_1d(sample, True)
    assert flag == 5, test_flags[flag]


def test_arr_to_col():
    """Test arr_to_col."""
    sample = np.asarray(np.random.normal(size=100), dtype=np.float64, order='F')
    flag = carma.arr_to_col(sample, False)
    assert flag == 0, test_flags[flag]


def test_arr_to_col_small():
    """Test arr_to_col."""
    sample = np.asarray(np.random.normal(size=5), dtype=np.float64, order='F')
    flag = carma.arr_to_col(sample, False)
    assert flag == 0, test_flags[flag]


def test_arr_to_col_2d():
    """Test arr_to_col."""
    sample = np.asarray(
        np.random.normal(size=(100, 1)), dtype=np.float64, order='F'
    )
    flag = carma.arr_to_col(sample, False)
    assert flag == 0, test_flags[flag]


def test_arr_to_col_C():
    """Test arr_to_col."""
    sample = np.asarray(np.random.normal(size=100), dtype=np.float64, order='C')
    flag = carma.arr_to_col(sample, False)
    assert flag == 0, test_flags[flag]


def test_arr_to_col_writeable_copy():
    """Test arr_to_col."""
    sample = np.asarray(np.random.normal(size=100), dtype=np.float64, order='F')
    sample.setflags(write=0)
    flag = carma.arr_to_col(sample, True)
    assert flag == 5, test_flags[flag]

def test_arr_to_col_copy():
    """Test arr_to_col."""
    sample = np.asarray(np.random.normal(size=100), dtype=np.float64, order='F')
    flag = carma.arr_to_col(sample, True)
    assert flag == 5, test_flags[flag]


def test_arr_to_col_copy_C():
    """Test arr_to_col."""
    sample = np.asarray(np.random.normal(size=100), dtype=np.float64, order='C')
    flag = carma.arr_to_col(sample, True)
    assert flag == 5, test_flags[flag]


def test_arr_to_row():
    """Test arr_to_row."""
    sample = np.asarray(np.random.normal(size=100), dtype=np.float64, order='F')
    flag = carma.arr_to_row(sample, False)
    assert flag == 0, test_flags[flag]


def test_arr_to_row_small():
    """Test arr_to_row."""
    sample = np.asarray(np.random.normal(size=5), dtype=np.float64, order='F')
    flag = carma.arr_to_row(sample, False)
    assert flag == 0, test_flags[flag]


def test_arr_to_row_2d():
    """Test arr_to_row."""
    sample = np.asarray(
        np.random.normal(size=(1, 100)), dtype=np.float64, order='F'
    )
    flag = carma.arr_to_row(sample, False)
    assert flag == 0, test_flags[flag]


def test_arr_to_row_C():
    """Test arr_to_row."""
    sample = np.asarray(np.random.normal(size=100), dtype=np.float64, order='C')
    flag = carma.arr_to_row(sample, False)
    assert flag == 0, test_flags[flag]


def test_arr_to_row_writeable_copy():
    """Test arr_to_row."""
    sample = np.asarray(np.random.normal(size=100), dtype=np.float64, order='F')
    sample.setflags(write=0)
    flag = carma.arr_to_row(sample, True)
    assert flag == 5, test_flags[flag]

def test_arr_to_row_copy():
    """Test arr_to_col."""
    sample = np.asarray(np.random.normal(size=100), dtype=np.float64, order='F')
    flag = carma.arr_to_row(sample, True)
    assert flag == 5, test_flags[flag]


def test_arr_to_row_copy_C():
    """Test arr_to_col."""
    sample = np.asarray(np.random.normal(size=100), dtype=np.float64, order='C')
    flag = carma.arr_to_row(sample, True)
    assert flag == 5, test_flags[flag]


def test_arr_to_cube():
    """Test arr_to_cube."""
    sample = np.asarray(
        np.random.normal(size=(25, 2, 2)), dtype=np.float64, order='F'
    )
    flag = carma.arr_to_cube(sample, False)
    assert flag == 0, test_flags[flag]


def test_arr_to_cube_small():
    """Test arr_to_cube."""
    sample = np.asarray(
        np.random.normal(size=(2, 2, 2)), dtype=np.float64, order='F'
    )
    flag = carma.arr_to_cube(sample, False)
    assert flag == 0, test_flags[flag]


def test_arr_to_cube_double_c_contiguous():
    """Test arr_to_mat."""
    sample = np.asarray(np.random.normal(size=(25, 2, 2)), dtype=np.float64)
    flag = carma.arr_to_cube(sample, False)
    assert flag == 5, test_flags[flag]


def test_arr_to_cube_copy():
    """Test arr_to_cube."""
    sample = np.asarray(
        np.random.normal(size=(25, 2, 2)), dtype=np.float64, order='F'
    )
    flag = carma.arr_to_cube(sample, True)
    assert flag == 5, test_flags[flag]


def test_arr_to_cube_writeable():
    """Test arr_to_mat."""
    sample = np.asarray(
        np.random.normal(size=(25, 2, 2)), dtype=np.float64, order='C'
    )
    sample.setflags(write=0)
    with pytest.raises(RuntimeError):
        flag = carma.arr_to_cube(sample, False)

def test_arr_to_cube_writeable_copy():
    """Test arr_to_mat."""
    sample = np.asarray(
        np.random.normal(size=(25, 2, 2)), dtype=np.float64, order='F'
    )
    sample.setflags(write=0)
    flag = carma.arr_to_cube(sample, True)
    assert flag == 5, test_flags[flag]

def test_to_arma_mat():
    """Test private implementation of to_arma for matrix."""
    sample = np.asarray(
        np.random.normal(size=(50, 2)), dtype=np.float64, order='F'
    )
    flag = carma.to_arma_mat(sample, False)
    assert flag == 0, test_flags[flag]


def test_to_arma_cube():
    """Test private implementation of to_arma for matrix."""
    sample = np.asarray(
        np.random.normal(size=(25, 2, 2)), dtype=np.float64, order='F'
    )
    flag = carma.to_arma_cube(sample, False)
    assert flag == 0, test_flags[flag]


def test_to_arma_col():
    """Test private implementation of to_arma for matrix."""
    sample = np.asarray(
        np.random.normal(size=100), dtype=np.float64, order='F'
    )
    flag = carma.to_arma_col(sample, False)
    assert flag == 0, test_flags[flag]


def test_to_arma_row():
    """Test private implementation of to_arma for matrix."""
    sample = np.asarray(
        np.random.normal(size=100), dtype=np.float64, order='F'
    )
    flag = carma.to_arma_row(sample, False)
    assert flag == 0, test_flags[flag]



import numpy as np
import test_carma as carma

# A canary test is a trivial test for testing the environment
def test_canary():
    print("Python running with")

    import site
    print('site.USER_SITE=', site.USER_SITE)

    import sys
    print('sys.path=', sys.path)

if __name__ == '__main__':
    test_canary()



"""Test numpy array to arma and back conversion functions."""
import numpy as np

import test_carma as carma


def test_mat_roundtrip():
    """Test mat_roundtrip."""
    og_sample = np.asarray(
        np.random.normal(size=(50, 2)), dtype=np.float64, order='F'
    )
    sample = og_sample.copy()
    out = carma.mat_roundtrip(sample)
    assert np.allclose(og_sample, out)

def test_mat_roundtrip_large():
    """Test mat_roundtrip."""
    og_sample = np.asarray(
        np.random.normal(size=(1000, 1000)), dtype=np.float64, order='F'
    )
    sample = og_sample.copy()
    out = carma.mat_roundtrip(sample)
    assert np.allclose(og_sample, out)


def test_mat_roundtrip_small():
    """Test mat_roundtrip."""
    og_sample = np.asarray(
        np.random.normal(size=(3, 3)), dtype=np.float64, order='F'
    )
    sample = og_sample.copy()
    out = carma.mat_roundtrip(sample)
    assert np.allclose(og_sample, out)


def test_mat_roundtrip_c_order():
    """Test mat_roundtrip."""
    og_sample = np.asarray(
        np.random.normal(size=(50, 2)), dtype=np.float64, order='C'
    )
    sample = og_sample.copy()
    out = carma.mat_roundtrip(sample)
    assert np.allclose(og_sample, out)

def test_mat_roundtrip_large_c_order():
    """Test mat_roundtrip."""
    og_sample = np.asarray(
        np.random.normal(size=(1000, 1000)), dtype=np.float64, order='C'
    )
    sample = og_sample.copy()
    out = carma.mat_roundtrip(sample)
    assert np.allclose(og_sample, out)


def test_mat_roundtrip_small_c_order():
    """Test mat_roundtrip."""
    og_sample = np.asarray(
        np.random.normal(size=(3, 3)), dtype=np.float64, order='C'
    )
    sample = og_sample.copy()
    out = carma.mat_roundtrip(sample)
    assert np.allclose(og_sample, out)


def test_row_roundtrip():
    """Test row_roundtrip."""
    og_sample = np.asarray(
        np.random.normal(size=(50)), dtype=np.float64, order='F'
    )
    sample = og_sample.copy()
    out = carma.row_roundtrip(sample)
    assert np.allclose(og_sample, out)

def test_row_roundtrip_large():
    """Test row_roundtrip."""
    og_sample = np.asarray(
        np.random.normal(size=(1000)), dtype=np.float64, order='F'
    )
    sample = og_sample.copy()
    out = carma.row_roundtrip(sample)
    assert np.allclose(og_sample, out)


def test_row_roundtrip_small():
    """Test row_roundtrip."""
    og_sample = np.asarray(
        np.random.normal(size=(3)), dtype=np.float64, order='F'
    )
    sample = og_sample.copy()
    out = carma.row_roundtrip(sample)
    assert np.allclose(og_sample, out)


def test_col_roundtrip():
    """Test col_roundtrip."""
    og_sample = np.asarray(
        np.random.normal(size=(50)), dtype=np.float64, order='F'
    )
    sample = og_sample.copy()
    out = carma.col_roundtrip(sample)
    assert np.allclose(og_sample, out.ravel())

def test_col_roundtrip_large():
    """Test col_roundtrip."""
    og_sample = np.asarray(
        np.random.normal(size=(1000)), dtype=np.float64, order='F'
    )
    sample = og_sample.copy()
    out = carma.col_roundtrip(sample)
    assert np.allclose(og_sample, out.ravel())


def test_col_roundtrip_small():
    """Test col_roundtrip."""
    og_sample = np.asarray(
        np.random.normal(size=(3)), dtype=np.float64, order='F'
    )
    sample = og_sample.copy()
    out = carma.col_roundtrip(sample)
    assert np.allclose(og_sample, out.ravel())


def test_cube_roundtrip():
    """Test cube_roundtrip."""
    og_sample = np.asarray(
        np.random.normal(size=(50, 3, 2)), dtype=np.float64, order='F'
    )
    sample = og_sample.copy()
    out = carma.cube_roundtrip(sample)
    assert np.allclose(og_sample, out)

def test_cube_roundtrip_large():
    """Test cube_roundtrip."""
    og_sample = np.asarray(
        np.random.normal(size=(1000, 10, 2)), dtype=np.float64, order='F'
    )
    sample = og_sample.copy()
    out = carma.cube_roundtrip(sample)
    assert np.allclose(og_sample, out)


def test_cube_roundtrip_small():
    """Test cube_roundtrip."""
    og_sample = np.asarray(
        np.random.normal(size=((2, 2, 2))), dtype=np.float64, order='F'
    )
    sample = og_sample.copy()
    out = carma.cube_roundtrip(sample)
    assert np.allclose(og_sample, out)



"""Test nparray.h."""
import pytest
import numpy as np

import test_carma as carma


def test_is_f_contiguous():
    """Test is_f_contiguous."""
    m = 'F order array should be F contiguous'
    sample = np.ones((10, 2), dtype=np.float64, order='F')
    assert carma.is_f_contiguous(sample) is True, m

    m = 'C order array should not be F contiguous'
    sample = np.ones((10, 2), dtype=np.float64, order='C')
    assert carma.is_f_contiguous(sample) is False, m


def test_is_c_contiguous():
    """Test is_c_contiguous."""
    m = 'C order array should be C contiguous'
    sample = np.ones((10, 2), dtype=np.float64, order='C')
    assert carma.is_c_contiguous(sample) is True, m

    m = 'F order array should not be C contiguous'
    sample = np.ones((10, 2), dtype=np.float64, order='F')
    assert carma.is_c_contiguous(sample) is False, m


def test_is_writeable():
    """Test is_writeable."""
    m = 'Array should be writeable'
    sample = np.ones((10, 2), dtype=np.float64, order='F')
    assert carma.is_writeable(sample) is True, m

    m = 'Array should not be writeable'
    sample.setflags(write=0)
    assert carma.is_writeable(sample) is False, m


def test_is_owndata():
    """Test is_writable."""
    m = 'Array should own the data'
    sample = np.ones((10, 2), dtype=np.float64, order='F')
    assert carma.is_owndata(sample) == sample.flags['OWNDATA'], m

    m = 'Array should not own the data'
    view = sample.reshape(20, 1)
    assert carma.is_owndata(view) == view.flags['OWNDATA'], m


def test_is_aligned():
    """Test is_aligned."""
    m = 'Array should be aligned'
    sample = np.arange(200, dtype=np.float64)
    assert carma.is_aligned(sample) == sample.flags['ALIGNED'], m

    m = 'Array should not be aligned'
    alt = np.frombuffer(sample.data, offset=2, count=100, dtype=np.float64)
    alt.shape = 10, 10
    assert carma.is_aligned(alt) == alt.flags['ALIGNED'], m


def test_set_not_owndata():
    """Test set_not_owndata."""
    m = 'Array should have owndata is false'
    sample = np.arange(100, dtype=np.float64)
    carma.set_not_owndata(sample)
    assert sample.flags['OWNDATA'] is False, m


def test_set_not_writeable():
    """Test set_not_writeable."""
    m = 'Array should have writeable is false'
    sample = np.arange(100, dtype=np.float64)
    carma.set_not_writeable(sample)
    assert sample.flags['WRITEABLE'] is False, m


def test_exception_flow():
    """Test exceptions are raised properly."""
    with pytest.raises(RuntimeError):
        carma.test_exception_flow()


def test_conversion_error():
    """Test exceptions are raised properly."""
    with pytest.raises(RuntimeError):
        carma.test_conversion_error()



"""Test armadillo matrix to  numpy array functions."""
import numpy as np

import test_carma as carma

###############################################################################
#                                 MAT
###############################################################################


def test_mat_to_arr():
    """Test mat_to_arr."""
    arr = carma.mat_to_arr(False)

    nan_cnt = np.isnan(arr).sum()
    assert nan_cnt == 0, 'Conversion introduced NaNs'

    zero_cnt = (arr == 0.0).sum()
    assert zero_cnt == 0, 'Conversion introduced zeros'

    inf_cnt = (arr == np.inf).sum()
    assert inf_cnt == 0, 'Conversion introduced inf values'

    near_null = (np.abs(arr) <= 1e-16).sum()
    assert near_null == 0, 'Conversion introduced near zero values'


def test_mat_to_arr_copy():
    """Test mat_to_arr."""
    arr = carma.mat_to_arr(True)

    nan_cnt = np.isnan(arr).sum()
    assert nan_cnt == 0, 'Conversion introduced NaNs'

    zero_cnt = (arr == 0.0).sum()
    assert zero_cnt == 0, 'Conversion introduced zeros'

    inf_cnt = (arr == np.inf).sum()
    assert inf_cnt == 0, 'Conversion introduced inf values'

    near_null = (np.abs(arr) <= 1e-16).sum()
    assert near_null == 0, 'Conversion introduced near zero values'


def test_mat_to_arr_plus_one():
    """Test mat_to_arr with c++ addition."""
    sample = np.asarray(
        np.random.normal(size=(20, 2)),
        dtype=np.float64,
        order='F'
    )
    mat = carma.mat_to_arr_plus_one(sample, False)
    assert np.allclose(mat, sample + 1)


def test_mat_to_arr_plus_one_copy():
    """Test mat_to_arr with c++ addition."""
    sample = np.asarray(
        np.random.normal(size=(20, 2)),
        dtype=np.float64,
        order='F'
    )
    mat = carma.mat_to_arr_plus_one(sample, True)
    assert np.allclose(mat, sample + 1)


def test_to_numpy_mat():
    arr = carma.to_numpy_mat(False)
    assert arr.ndim == 2
    assert arr.shape[0] == 4
    assert arr.shape[1] == 5
    assert arr.flags['WRITEABLE'] == True

    arr = carma.to_numpy_mat(True)
    assert arr.ndim == 2
    assert arr.shape[0] == 4
    assert arr.shape[1] == 5
    assert arr.flags['WRITEABLE'] == True


def test_to_numpy_view_mat():
    arr = carma.to_numpy_view_mat()
    assert arr.ndim == 2
    assert arr.shape[0] == 4
    assert arr.shape[1] == 5
    assert arr.flags['WRITEABLE'] == False


###############################################################################
#                                 ROW
###############################################################################


def test_row_to_arr():
    """Test row_to_arr."""
    arr = carma.row_to_arr(False)

    nan_cnt = np.isnan(arr).sum()
    assert nan_cnt == 0, 'Conversion introduced NaNs'

    zero_cnt = (arr == 0.0).sum()
    assert zero_cnt == 0, 'Conversion introduced zeros'

    inf_cnt = (arr == np.inf).sum()
    assert inf_cnt == 0, 'Conversion introduced inf values'

    near_null = (np.abs(arr) <= 1e-16).sum()
    assert near_null == 0, 'Conversion introduced near zero values'


def test_row_to_arr_copy():
    """Test row_to_arr."""
    arr = carma.row_to_arr(True)

    nan_cnt = np.isnan(arr).sum()
    assert nan_cnt == 0, 'Conversion introduced NaNs'

    zero_cnt = (arr == 0.0).sum()
    assert zero_cnt == 0, 'Conversion introduced zeros'

    inf_cnt = (arr == np.inf).sum()
    assert inf_cnt == 0, 'Conversion introduced inf values'

    near_null = (np.abs(arr) <= 1e-16).sum()
    assert near_null == 0, 'Conversion introduced near zero values'


def test_row_to_arr_plus_one():
    """Test row_to_arr with c++ addition."""
    sample = np.asarray(
        np.random.normal(size=(20)),
        dtype=np.float64,
        order='F'
    )
    mat = carma.row_to_arr_plus_one(sample, False)
    assert np.allclose(mat, sample + 1)


def test_row_to_arr_plus_one_copy():
    """Test row_to_arr with c++ addition."""
    sample = np.asarray(
        np.random.normal(size=(20)),
        dtype=np.float64,
        order='F'
    )
    mat = carma.row_to_arr_plus_one(sample, True)
    assert np.allclose(mat, sample + 1)


def test_to_numpy_row():
    arr = carma.to_numpy_row(False)
    assert arr.ndim == 2
    assert arr.shape[0] == 1
    assert arr.shape[1] == 20
    assert arr.flags['WRITEABLE'] == True

    arr = carma.to_numpy_row(True)
    assert arr.ndim == 2
    assert arr.shape[0] == 1
    assert arr.shape[1] == 20
    assert arr.flags['WRITEABLE'] == True


def test_to_numpy_view_row():
    arr = carma.to_numpy_view_row()
    assert arr.ndim == 2
    assert arr.shape[0] == 1
    assert arr.shape[1] == 20
    assert arr.flags['WRITEABLE'] == False


###############################################################################
#                                 COL
###############################################################################


def test_col_to_arr():
    """Test col_to_arr."""
    arr = carma.col_to_arr(False)

    nan_cnt = np.isnan(arr).sum()
    assert nan_cnt == 0, 'Conversion introduced NaNs'

    zero_cnt = (arr == 0.0).sum()
    assert zero_cnt == 0, 'Conversion introduced zeros'

    inf_cnt = (arr == np.inf).sum()
    assert inf_cnt == 0, 'Conversion introduced inf values'

    near_null = (np.abs(arr) <= 1e-16).sum()
    assert near_null == 0, 'Conversion introduced near zero values'


def test_col_to_arr_copy():
    """Test col_to_arr."""
    arr = carma.col_to_arr(True)

    nan_cnt = np.isnan(arr).sum()
    assert nan_cnt == 0, 'Conversion introduced NaNs'

    zero_cnt = (arr == 0.0).sum()
    assert zero_cnt == 0, 'Conversion introduced zeros'

    inf_cnt = (arr == np.inf).sum()
    assert inf_cnt == 0, 'Conversion introduced inf values'

    near_null = (np.abs(arr) <= 1e-16).sum()
    assert near_null == 0, 'Conversion introduced near zero values'


def test_col_to_arr_plus_one():
    """Test col_to_arr with c++ addition."""
    sample = np.asarray(
        np.random.normal(size=(20, 1)),
        dtype=np.float64,
        order='F'
    )
    mat = carma.col_to_arr_plus_one(sample, False)
    assert np.allclose(mat, sample + 1)


def test_col_to_arr_plus_one_copy():
    """Test col_to_arr with c++ addition."""
    sample = np.asarray(
        np.random.normal(size=(20, 1)),
        dtype=np.float64,
        order='F'
    )
    mat = carma.col_to_arr_plus_one(sample, True)
    assert np.allclose(mat, sample + 1)


def test_to_numpy_col():
    arr = carma.to_numpy_col(False)
    assert arr.shape[0] == 20
    assert arr.shape[1] == 1
    assert arr.flags['WRITEABLE'] == True

    arr = carma.to_numpy_col(True)
    assert arr.shape[0] == 20
    assert arr.shape[1] == 1
    assert arr.flags['WRITEABLE'] == True


def test_to_numpy_view_col():
    arr = carma.to_numpy_view_col()
    assert arr.ndim == 2
    assert arr.shape[0] == 20
    assert arr.shape[1] == 1
    assert arr.flags['WRITEABLE'] == False


###############################################################################
#                                 CUBE
###############################################################################


def test_cube_to_arr():
    """Test cube_to_arr."""
    arr = carma.cube_to_arr(False)

    nan_cnt = np.isnan(arr).sum()
    assert nan_cnt == 0, 'Conversion introduced NaNs'

    zero_cnt = (arr == 0.0).sum()
    assert zero_cnt == 0, 'Conversion introduced zeros'

    inf_cnt = (arr == np.inf).sum()
    assert inf_cnt == 0, 'Conversion introduced inf values'

    near_null = (np.abs(arr) <= 1e-16).sum()
    assert near_null == 0, 'Conversion introduced near zero values'


def test_cube_to_arr_return():
    """Test cube_to_arr."""
    arr = carma.cube_to_arr(True)

    nan_cnt = np.isnan(arr).sum()
    assert nan_cnt == 0, 'Conversion introduced NaNs'

    zero_cnt = (arr == 0.0).sum()
    assert zero_cnt == 0, 'Conversion introduced zeros'

    inf_cnt = (arr == np.inf).sum()
    assert inf_cnt == 0, 'Conversion introduced inf values'

    near_null = (np.abs(arr) <= 1e-6).sum()
    assert near_null == 0, 'Conversion introduced near zero values'


def test_cube_to_arr_plus_one():
    """Test cube_to_arr with c++ addition."""
    or_sample = np.random.normal(size=(20, 3, 2))
    sample = np.asarray(
        or_sample,
        dtype=np.float64,
        order='F'
    )
    mat = carma.cube_to_arr_plus_one(sample, False)
    assert np.allclose(mat, 1 + sample)


def test_cube_to_arr_plus_one_copy():
    """Test cube_to_arr with c++ addition."""
    or_sample = np.random.normal(size=(20, 3, 2))
    sample = np.asarray(
        or_sample,
        dtype=np.float64,
        order='F'
    )
    mat = carma.cube_to_arr_plus_one(sample, True)
    assert np.allclose(mat, 1 + sample)


def test_to_numpy_cube():
    arr = carma.to_numpy_cube(False)
    assert arr.ndim == 3
    assert arr.shape[0] == 100
    assert arr.shape[1] == 2
    assert arr.shape[2] == 4
    assert arr.flags['WRITEABLE'] == True

    arr = carma.to_numpy_cube(True)
    assert arr.ndim == 3
    assert arr.shape[0] == 100
    assert arr.shape[1] == 2
    assert arr.shape[2] == 4
    assert arr.flags['WRITEABLE'] == True


def test_to_numpy_view_cube():
    arr = carma.to_numpy_view_cube()
    assert arr.ndim == 3
    assert arr.shape[0] == 100
    assert arr.shape[1] == 2
    assert arr.shape[2] == 4
    assert arr.flags['WRITEABLE'] == False



import platform
import sys
import os
import tempfile
import setuptools
import subprocess
from setuptools import setup, Extension
from setuptools.command.build_ext import build_ext
import distutils.sysconfig
import distutils.spawn

__version__ = "5.0.0"

# TODO(@motiwari): Move this to a separate file
GHA = "GITHUB_ACTIONS"


class get_pybind_include(object):
    """
    Helper class to determine the pybind11 include path.

    The purpose of this class is to postpone importing pybind11
    until it is actually installed via setup's setup_requires arg,
    so that the ``get_include()`` method can be invoked.
    """

    def __str__(self):
        import pybind11

        return pybind11.get_include()


class get_numpy_include(object):
    """
    Helper class to determine the numpy include path
    The purpose of this class is to postpone importing numpy
    until it is actually installed via setup's setup_requires arg,
    so that the ``get_include()`` method can be invoked.
    """

    def __str__(self):
        import numpy

        return numpy.get_include()


def compiler_check():
    """
    This is necessary because setuptools will use the compiler that compiled
    python for some of the compilation process, even if the user specifies
    another one!
    """
    if sys.platform == "darwin":
        # Note: On Mac, we should always use LLVM clang
        # because Apple's clang does not support OpenMP
        # and we need OpenMP to parallelize the code
        # Note: distutils.sysconfig.get_config_vars() works for Python 3.11+, but we need to use os.environ for Python 3.10
        if "clang" not in distutils.sysconfig.get_config_vars()["CC"] and "clang" not in dict(os.environ)["CC"]:
            raise Exception(
                "Need to install LLVM clang! \
                Please run `brew install llvm`"
            )
        return "clang"

    # For all platforms
    try:
        return (
            "clang"
            if "clang" in distutils.sysconfig.get_config_vars()["CC"]
            else "gcc"
        )
    except KeyError:
        # The 'CC' environment variable hasn't been set.
        # In this case, search for compilers that we can use.
        # Borrowed from https://github.com/clab/dynet/blob/master/setup.py
        if distutils.spawn.find_executable("cl") is not None:
            return "msvc"
        elif distutils.spawn.find_executable("clang") is not None:
            return "clang"
        elif distutils.spawn.find_executable("gcc") is not None:
            return "gcc"

    raise Exception(
        "No C++ compiler was found. Please ensure you have "
        "MSVC, LLVM clang, or GCC."
    )


def has_flag(compiler: str, flagname: str):
    """
    Return a boolean indicating whether a flag name is supported on the
    specified compiler.
    """
    with tempfile.NamedTemporaryFile("w", suffix=".cpp", delete=False) as f:
        f.write("int main (int argc, char **argv) { return 0; }")
        fname = f.name
    try:
        compiler.compile([fname], extra_postargs=[flagname])
    except setuptools.distutils.errors.CompileError:
        return False
    finally:
        try:
            os.remove(fname)
        except OSError:
            print("Warning: Received an OSError")
    return True


def cpp_flag(compiler: str):
    """
    Return the -std=c++[11/14/17] compiler flag.

    The newer version is prefered over c++11 (when it is available).
    """
    compiler_name = compiler_check()
    if compiler_name == "clang" or compiler_name == "gcc":
        flags = ["-std=c++17"]
    else:
        # Assume msvc
        flags = ["/std:c++17"]  # required for std::optional

    for flag in flags:
        if has_flag(compiler, flag):
            return flag

    raise RuntimeError(
        "Unsupported compiler -- \
        at least C++11 support is needed!"
    )


def check_brew_package(pkg_name: str):
    brew_cmd = ["brew", "--prefix", pkg_name]
    process = subprocess.Popen(brew_cmd, stdout=subprocess.PIPE)
    output, _error = process.communicate()
    if output.decode() == "":
        raise Exception(
            "Error: Need to install %s via homebrew! \
            Please run `brew install %s`"
            % (pkg_name, pkg_name)
        )

    return output.decode().strip()


def check_brew_installation():
    brew_cmd = ["which", "brew"]
    process = subprocess.Popen(brew_cmd, stdout=subprocess.PIPE)
    output, _error = process.communicate()
    if output.decode() == "":
        raise Exception(
            "Error: Need to install homebrew! \
            Please see https://brew.sh"
        )


def check_numpy_installation():
    try:
        import numpy  # noqa: F401
    except ModuleNotFoundError:
        raise Exception("Need to install numpy!")


def install_check_mac():
    # Make sure homebrew is installed
    check_brew_installation()

    # Make sure numpy is installed
    check_numpy_installation()

    # Check that libomp, armadillo, and openblas are installed
    check_brew_package("libomp")
    check_brew_package("armadillo")
    check_brew_package("openblas")

    # Because we don't want to force M1 users to compile from source,
    # we build wheels on Github Runners via Github Actions. In
    # doing so, we need to use Apple's compiler to cross-compile
    # for M1 Macs and should NOT use clang.
    # If we are building from source on a non-M1 Mac, we should use
    # LLVM clang to support multithreading via OpenMP
    # TODO(@motiwari): Check if the arm64 wheels are not multithreaded
    # TODO(@motiwari): Check if the universal2 wheels are not multithreaded
    # when installed on Intel Mac
    if not os.environ.get(GHA, False):
        # If we are NOT running inside a Github action,
        # check that LLVM clang is installed
        llvm_loc = check_brew_package(
            "llvm"
        )  # We need to use LLVM clang since Apple's doesn't support OpenMP

        # Set compiler to LLVM clang on Mac for OpenMP support
        distutils.sysconfig.get_config_vars()["CC"] = os.path.join(
            llvm_loc, "bin", "clang"
        )


def check_omp_install_linux():
    # TODO: Need to get exact compiler name and version to check this
    # Check compiler version is gcc>=6.0.0 or clang>=X.X.X
    pass


def check_armadillo_install_linux():
    # Since armadillo is a C++ extension, just check if it exists
    cmd = ["find", "/", "-iname", "armadillo"]
    process = subprocess.Popen(
        cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.DEVNULL,
    )
    output, _error = process.communicate()
    if output.decode() == "":
        print(
            "Warning: Armadillo may not be installed. \
            Please build it from",
            os.path.join(
                "BanditPAM",
                "headers",
                "carma",
                "third_party",
                "armadillo-code",
            ),
        )

    return output.decode().strip()


def check_linux_package_installation(pkg_name: str):
    cmd = ["dpkg", "-s", pkg_name]
    process = subprocess.Popen(cmd, stdout=subprocess.PIPE)
    output, _error = process.communicate()
    if output.decode() == "":
        raise Exception(
            "Error: Need to install %s! \
            Please ensure all dependencies are installed \
            via your package manager (apt, yum, etc.): \
            build-essential checkinstall \
            libssl-dev libsqlite3-dev tk-dev \
            libgdbm-dev libc6-dev libbz2-dev libffi-dev zlib1g-dev"
            % (pkg_name)
        )

    return output.decode().strip()


def install_ubuntu_pkgs():
    # TODO: Remove dangerous os.system() calls
    # See https://stackoverflow.com/a/51329156
    os.system("apt update")
    os.system(
        "apt install -y \
        build-essential \
        checkinstall \
        libssl-dev \
        libsqlite3-dev \
        libgdbm-dev \
        libc6-dev \
        libbz2-dev \
        libffi-dev \
        zlib1g-dev"
    )  # noqa: E124


def setup_colab(delete_source=False):
    # If we're in Google Colab, we need to manually copy over
    # the prebuilt armadillo libraries
    # NOTE: This only works for Colab instances with Ubuntu 18.04 Runtimes!
    try:
        import google.colab  # noqa: F401

        in_colab = True
    except ModuleNotFoundError:
        in_colab = False

    if in_colab:
        # TODO: Remove dangerous os.system() calls
        # See https://stackoverflow.com/a/51329156
        install_ubuntu_pkgs()

        # TODO(@motiwari): Make this a randomly-named directory
        # and set delete_source=True always
        repo_location = os.path.join("/", "content", "BanditPAM")
        # Note the space after the git URL to separate the source and target
        os.system(
            "git clone https://github.com/motiwari/BanditPAM.git "
            + repo_location
        )
        os.system(
            repo_location
            + os.path.join(
                "scripts",
                "colab_files",
                "colab_install_armadillo.sh",
            )
        )
        if delete_source:
            os.system("rm -rf " + repo_location)


def setup_paperspace_gradient():
    # Unfortunately, Paperspace Gradient does not make it easy to tell
    # whether you're inside a Gradient instance. For this reason, we
    # determine whether python is running inside a Gradient instance
    # by checking for a /notebooks directory

    # TODO: Remove dangerous os.system() calls
    # See https://stackoverflow.com/a/51329156

    in_paperspace_gradient = os.path.exists("/notebooks")

    if in_paperspace_gradient:
        install_ubuntu_pkgs()
        os.system(
            "DEBIAN_FRONTEND=noninteractive TZ=America/NewYork \
            apt install -y tk-dev"
        )
        os.system(
            "git clone \
            https://gitlab.com/conradsnicta/armadillo-code.git"
        )
        os.system("cd armadillo-code && cmake . && make install")


def install_check_ubuntu():
    # Make sure linux packages are installed
    dependencies = [
        "build-essential",
        "checkinstall",
        "libssl-dev",
        "libsqlite3-dev",
        "tk-dev",
        "libgdbm-dev",
        "libc6-dev",
        "libbz2-dev",
        "libffi-dev",
        "zlib1g-dev",
    ]

    setup_colab(delete_source=False)

    setup_paperspace_gradient()

    for dep in dependencies:
        check_linux_package_installation(dep)

    # Make sure numpy is installed
    check_numpy_installation()

    # Check openMP is installed
    check_omp_install_linux()

    # Check armadillo is installed
    check_armadillo_install_linux()


def is_ubuntu():
    cmd = ["cat", os.path.join("/", "etc", "issue")]
    process = subprocess.Popen(cmd, stdout=subprocess.PIPE)
    output, _error = process.communicate()
    return "Ubuntu" in output.decode()


def get_package_prefix(package=None):
    assert package is not None, "Package name must be provided"
    try:
        result = subprocess.run(
            ["brew", "--prefix", package],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            check=True,
        )
        return result.stdout.strip().decode()
    except subprocess.CalledProcessError as e:
        print("Error running brew:", e.stderr)
        return None


class BuildExt(build_ext):
    """
    A custom build extension for adding compiler-specific options.
    """

    c_opts = {"msvc": ["/EHsc"], "unix": []}
    l_opts = {"msvc": [], "unix": []}

    if sys.platform == "darwin":
        install_check_mac()
        # Verify that we're either compiling with clang or
        # inside a Github Action
        assert compiler_check() == "clang" or os.environ.get(
            GHA, False
        ), "Need to install LLVM clang!"
        darwin_opts = [
            "-stdlib=libc++",
            "-mmacosx-version-min=10.14",
            "-O3",
        ]
        c_opts["unix"] += darwin_opts
        l_opts["unix"] += darwin_opts
    elif sys.platform == "linux" or sys.platform == "linux2":
        if is_ubuntu():
            install_check_ubuntu()

        linux_opts = ["-O3"]
        c_opts["unix"] += linux_opts
        l_opts["unix"] += linux_opts
    # Currently necessary (unsure why)
    elif sys.platform == "win32":
        c_opts["msvc"] += ["/fsanitize=address"]

    def build_extensions(self):
        ct = self.compiler.compiler_type

        comp_opts = self.c_opts.get(ct, [])
        link_opts = self.l_opts.get(ct, [])

        # TODO(@motiwari): on Windows, these flags are unrecognized
        comp_opts.append(cpp_flag(self.compiler))
        comp_opts.append("-O3")

        if sys.platform == "darwin":
            for package in ["armadillo", "libomp", "openblas"]:
                package_prefix = get_package_prefix(package)
                link_opts.append(f"-Wl,-rpath,{package_prefix}/lib")

            if os.environ.get(GHA, False):
                # We are inside a Github Runner on a Mac.
                comp_opts.append("-Xpreprocessor")  # NEEDS TO BE WITH NEXT LINE
                comp_opts.append("-fopenmp")  # NEEDS TO BE WITH PREVIOUS LINE

                comp_opts.append("-lomp")  # Potentially unused?
                comp_opts.append("-lopenblas")  # Potentially unused?

                # TODO(@motiwari): include armadillo here?

                for package in ["libomp", "openblas"]:
                    package_prefix = get_package_prefix(package)
                    comp_opts.append("-I{}/include".format(package_prefix))
                    comp_opts.append("-L{}/lib".format(package_prefix))
        elif sys.platform != "win32":
            comp_opts.append("-fopenmp")

        compiler_name = compiler_check()
        if sys.platform == "darwin":
            if compiler_name == "gcc":
                link_opts.append("-lomp")
            else:
                link_opts.append("-lomp")
        elif sys.platform == "linux" or sys.platform == "linux2":
            if compiler_name == "gcc":
                link_opts.append("-lgomp")
            else:
                # Both clang and Apple's clang should use libomp
                link_opts.append("-lomp")
        else:
            # On Windows
            # TODO(@motiwari): Fix this
            pass

        if ct == "unix":
            if has_flag(self.compiler, "-fvisibility=hidden"):
                comp_opts.append("-fvisibility=hidden")

        for ext in self.extensions:
            ext.define_macros = [
                (
                    "VERSION_INFO",
                    '"{}"'.format(self.distribution.get_version()),
                )
            ]
            ext.extra_compile_args = comp_opts
            ext.extra_compile_args += []  # []["-arch", "x86_64"]

            ext.extra_link_args = link_opts
            ext.extra_link_args += [
                "-v",
            ]  # "-arch", "x86_64"]

        build_ext.build_extensions(self)


def main():
    if sys.platform == "linux" or sys.platform == "linux2":
        include_dirs = [
            get_pybind_include(),
            get_numpy_include(),
            "headers",
            os.path.join("headers", "algorithms"),
            os.path.join("headers", "python_bindings"),
            os.path.join("headers", "carma", "include"),
            os.path.join("headers", "carma", "include", "carma_bits"),
            os.path.join("/", "usr", "local", "include"),
        ]
    elif sys.platform == "darwin":  # OSX
        include_dirs = [
            get_pybind_include(),
            get_numpy_include(),
            "headers",
            os.path.join("headers", "algorithms"),
            os.path.join("headers", "python_bindings"),
            os.path.join("headers", "carma", "include"),
            os.path.join("headers", "carma", "include", "carma"),
            os.path.join("headers", "carma", "include", "carma", "carma"),
            # To include carma when the BanditPAM repo hasnt been initialized
            os.path.join("/", "usr", "local", "include"),
            os.path.join("/", "usr", "local", "include", "carma"),
            os.path.join(
                "/", "usr", "local", "include", "carma", "carma_bits"
            ),
            # When building from source on M1 Macs, may need these dirs
            # Currently, we should never be building from source on an M1 Mac,
            # Only cross-compiling from an Intel Mac
            # TODO(@motiwari): Remove extraneous directories
            os.path.join("/", "opt", "homebrew"),
            os.path.join("/", "opt", "homebrew", "bin"),
            os.path.join("/", "opt", "homebrew", "include"),
            os.path.join("/", "opt", "homebrew", "lib"),
            os.path.join("/", "opt", "homebrew", "opt"),
            os.path.join("/", "opt", "homebrew", "opt", "armadillo"),
            os.path.join("/", "opt", "homebrew", "opt", "openblas"),
            os.path.join(
                "/", "opt", "homebrew", "opt", "armadillo", "include"
            ),
            os.path.join(
                "/",
                "opt",
                "homebrew",
                "opt",
                "armadillo",
                "include",
                "armadillo_bits",
            ),
        ]
        for package in ["armadillo", "openblas", "libomp"]:
            package_prefix = get_package_prefix(package)
            include_dirs.append(os.path.join(package_prefix, "include"))
    elif sys.platform == "win32":  # WIN32
        include_dirs = [
            get_pybind_include(),
            get_numpy_include(),
            "headers",
            os.path.join("headers", "algorithms"),
            os.path.join("headers", "python_bindings"),
            os.path.join("headers", "carma", "include"),
            os.path.join("headers", "carma", "include", "carma_bits"),
            os.path.join("headers", "armadillo", "include"),
            os.path.join("headers", "armadillo", "include", "armadillo_bits"),
        ]
    else:
        raise Exception("Unrecognized platform")

    compiler_name = compiler_check()
    if sys.platform == "darwin" and os.environ.get(GHA, False):
        # On Mac Github Runners, we should NOT include gomp or omp here
        # due to build errors.
        libraries = ["armadillo", "openblas"]
    elif sys.platform == "win32":
        libraries = ["libopenblas"]
    else:
        if compiler_name == "gcc":
            # TODO(@motiwari): Perhaps change this to c++ when we also update cpp_args below
            libraries = ["armadillo", "gomp", "stdc++"]
        else:  # clang
            libraries = ["armadillo", "omp", "stdc++"]

    if sys.platform == "win32":
        cpp_args = ["/std:c++17"]
        library_dirs = [
            # for windows
            os.path.join(os.getcwd(), r"headers\armadillo\examples\lib_win64"),
        ]
    else:
        cpp_args = [
            "-static-libstdc++"
        ]  # TODO(@motiwari): Modify this based on gcc or clang
        library_dirs = [
            os.path.join("/", "usr", "local", "lib"),
        ]
        if sys.platform == "darwin":
            for package in ["armadillo", "libomp", "openblas"]:
                package_prefix = get_package_prefix(package)
                library_dirs.append(os.path.join(package_prefix, "lib"))

            if platform.processor() == "arm":  # M1
                library_dirs.append(
                    os.path.join("/", "opt", "homebrew", "opt", "armadillo", "lib")
                )
                library_dirs.append(
                    os.path.join("/", "opt", "homebrew", "opt", "libomp", "lib")
                )
                library_dirs.append(
                    os.path.join("/", "opt", "homebrew", "opt", "openblas", "lib")
                )

    ext_modules = [
        Extension(
            "banditpam",
            [
                os.path.join("src", "algorithms", "kmedoids_algorithm.cpp"),
                os.path.join("src", "algorithms", "pam.cpp"),
                os.path.join("src", "algorithms", "banditpam.cpp"),
                os.path.join("src", "algorithms", "banditpam_orig.cpp"),
                os.path.join("src", "algorithms", "fastpam1.cpp"),
                os.path.join(
                    "src", "python_bindings", "kmedoids_pywrapper.cpp"
                ),
                os.path.join("src", "python_bindings", "medoids_python.cpp"),
                os.path.join(
                    "src", "python_bindings", "build_medoids_python.cpp"
                ),
                os.path.join("src", "python_bindings", "fit_python.cpp"),
                os.path.join("src", "python_bindings", "labels_python.cpp"),
                os.path.join("src", "python_bindings", "steps_python.cpp"),
                os.path.join("src", "python_bindings", "loss_python.cpp"),
                os.path.join("src", "python_bindings", "cache_python.cpp"),
                os.path.join(
                    "src", "python_bindings", "swap_times_python.cpp"
                ),
            ],
            include_dirs=include_dirs,
            library_dirs=library_dirs,
            libraries=libraries,
            language="c++1z",  # TODO: modify this based on cpp_flag(compiler)
            extra_compile_args=cpp_args,
            extra_link_args=["-vvvv"],  # Wrong pass (BuildExt sets extra_link_args)
        )
    ]

    with open(os.path.join("docs", "long_desc.rst"), encoding="utf-8") as f:
        long_description = f.read()

    my_data_files = [("docs", [os.path.join("docs", "long_desc.rst")])]
    if sys.platform == "win32":
        my_data_files.append(
            (
                "",
                [
                    os.path.join(
                        os.getcwd(),
                        r"headers\armadillo\examples\lib_win64"
                        + r"\libopenblas.dll",
                    )
                ],
            )
        )

    setup(
        name="banditpam",
        version=__version__,
        author="Mo Tiwari",
        maintainer="Mo Tiwari",
        author_email="motiwari@stanford.edu",
        url="https://github.com/motiwari/BanditPAM",
        description="BanditPAM: A state-of-the-art, \
            high-performance k-medoids algorithm.",
        long_description=long_description,
        ext_modules=ext_modules,
        setup_requires=["pybind11>=2.5.0", "numpy>=1.18"],
        data_files=my_data_files,
        include_package_data=True,
        cmdclass={"build_ext": BuildExt},
        zip_safe=False,
        classifiers=[
            "Programming Language :: Python :: 3",
            "License :: OSI Approved :: MIT License",
            "Operating System :: OS Independent",
        ],
        headers=[
            os.path.join("headers", "algorithms", "kmedoids_algorithm.hpp"),
            os.path.join("headers", "algorithms", "banditpam.hpp"),
            os.path.join("headers", "algorithms", "fastpam1.hpp"),
            os.path.join("headers", "algorithms", "pam.hpp"),
            os.path.join(
                "headers", "python_bindings", "kmedoids_pywrapper.hpp"
            ),
        ],
    )


if __name__ == "__main__":
    main()



"""Tests for ArrayStore class."""
import pytest
import numpy as np
import test_carma as carma


def test_ArrayStore_get_view():
    """Tests for ArrayStore class.get_view()."""
    sample = np.asarray(np.random.uniform(-1, 1, size=100), order='F')
    arraystore = carma.dArrayStore(sample, True)
    arr = arraystore.get_view(False)
    assert arr.flags['OWNDATA'] is False
    np.testing.assert_allclose(arr.flatten(), sample)


def test_ArrayStore_copy():
    """Test ArrayStore class when not stealing."""
    og_sample = np.asarray(np.random.uniform(-1, 1, size=100), order='F')
    sample = og_sample.copy()

    arraystore = carma.dArrayStore(sample, True)
    arr = arraystore.get_view(False)
    np.testing.assert_allclose(arr.flatten(), og_sample)

    # trigger descructor
    arraystore = None
    del arraystore
    arr = None
    del arr
    # Validate the memory of sample is untouched
    assert np.allclose(sample, og_sample)


def test_ArrayStore_non_writeable():
    """Test ArrayStore class when marked as non-readable."""
    sample = np.asarray(np.random.uniform(-1, 1, size=100), order='F')
    arraystore = carma.dArrayStore(sample, True)
    arr = arraystore.get_view(False)
    assert arr.flags['OWNDATA'] is False
    assert arr.flags['WRITEABLE'] is False
    with pytest.raises(ValueError):
        arr[0, 0] = 1.0


def test_ArrayStore_writeable():
    """Test ArrayStore class when marked as writeable."""
    sample = np.asarray(np.random.uniform(-1, 1, size=100), order='F')
    arraystore = carma.dArrayStore(sample, True)
    arr = arraystore.get_view(True)
    assert arr.flags['OWNDATA'] is False
    assert arr.flags['WRITEABLE'] is True
    arr[0, 0] = 1.0


def test_ArrayStore_steal():
    """Test ArrayStore class when we steal the memory."""
    og_sample = np.asarray(np.random.uniform(-1, 1, size=100), order='F')
    sample = og_sample.copy()

    arraystore = carma.dArrayStore(sample, False)
    arr = arraystore.get_view(True)
    np.testing.assert_allclose(arr.flatten(), og_sample)

    # trigger destructor
    arraystore = None
    del arraystore
    arr = None
    del arr


def test_ArrayStore_set_data():
    """Test ArrayStore class function set_data."""
    sample1 = np.asarray(np.random.uniform(-1, 1, size=100), order='F')
    og_sample1 = np.copy(sample1)
    sample2 = np.asarray(np.random.uniform(-1, 1, size=100), order='F')
    og_sample2 = np.copy(sample2)

    arraystore = carma.dArrayStore(sample1, True)
    arr = arraystore.get_view(True)
    np.testing.assert_allclose(arr.flatten(), og_sample1)

    arraystore.set_array(sample2, False)
    arr = arraystore.get_view(True)
    np.testing.assert_allclose(arr.flatten(), og_sample2)
    assert arr.flags['OWNDATA'] is False
    assert arr.flags['WRITEABLE'] is True


def test_ArrayStore_set_data_flags():
    """Test ArrayStore class function set_data."""
    sample1 = np.asarray(np.random.uniform(-1, 1, size=100), order='F')
    sample2 = np.asarray(np.random.uniform(-1, 1, size=100), order='F')

    arraystore = carma.dArrayStore(sample1, True)
    arr = arraystore.get_view(True)
    assert arr.flags['OWNDATA'] is False
    assert arr.flags['WRITEABLE'] is True

    arraystore.set_array(sample2, True)
    arr = arraystore.get_view(False)
    assert arr.flags['OWNDATA'] is False
    assert arr.flags['WRITEABLE'] is False


def test_ArrayStore_get_view_float():
    """Tests for ArrayStore class.get_view()."""
    sample = np.asarray(
        np.random.uniform(-1, 1, size=100),
        order='F',
        dtype=np.float32
    )
    arraystore = carma.fArrayStore(sample, True)
    arr = arraystore.get_view(False)
    assert arr.flags['OWNDATA'] is False
    np.testing.assert_allclose(arr.flatten(), sample)


def test_ArrayStore_get_view_long():
    """Tests for ArrayStore class.get_view()."""
    sample = np.asarray(
        np.random.randint(-10, 10, size=100).astype(np.int64),
        order='F'
    )
    arraystore = carma.lArrayStore(sample, True)
    arr = arraystore.get_view(False)
    assert arr.flags['OWNDATA'] is False
    np.testing.assert_allclose(arr.flatten(), sample)


def test_ArrayStore_get_view_int():
    """Tests for ArrayStore class.get_view()."""
    sample = np.asarray(
        np.random.randint(-10, 10, size=100).astype(np.int32),
        order='F',
        dtype=np.int32
    )
    arraystore = carma.iArrayStore(sample, True)
    arr = arraystore.get_view(False)
    assert arr.flags['OWNDATA'] is False
    np.testing.assert_allclose(arr.flatten(), sample)


def test_ArrayStore_get_mat():
    """Tests for ArrayStore C++ api."""
    delta = carma.test_ArrayStore_get_mat()
    assert delta < 1e-6


def test_ArrayStore_get_mat_rvalue():
    """Tests for ArrayStore C++ api."""
    delta = carma.test_ArrayStore_get_mat()
    assert delta < 1e-6


def test_ArrayStore_get_view_cpp():
    """Tests for ArrayStore get_view."""
    arr = carma.test_ArrayStore_get_view(True)
    assert arr.flags['OWNDATA'] is False
    assert arr.flags['WRITEABLE'] is True

    arr = carma.test_ArrayStore_get_view(False)
    assert arr.flags['OWNDATA'] is False
    assert arr.flags['WRITEABLE'] is False



import unittest
import pandas as pd
import numpy as np
import sys

from banditpam import KMedoids
from utils import bpam_agrees_pam
from constants import (
    NUM_SMALL_CASES,
    SMALL_K_SCHEDULE,
    N_SMALL_K,
    SMALL_SAMPLE_SIZE,
    PROPORTION_PASSING,
)


class SmallerTests(unittest.TestCase):
    small_mnist = pd.read_csv("data/MNIST_100.csv", header=None).to_numpy()
    mnist_70k = pd.read_csv("data/MNIST_70k.csv", sep=" ", header=None)
    if sys.platform == "win32":
        scrna = pd.read_csv(
            "data/scrna_reformat.csv.gz", header=None, dtype="float16"
        )  # float16 for less memory usage
    else:
        scrna = pd.read_csv("data/scrna_reformat.csv.gz", header=None)

    def test_small_mnist(self):
        """
        Test NUM_SMALL_CASES number of test cases with subsets of size
        SMALL_SAMPLE_SIZE randomly drawn from the full MNIST dataset.
        """
        for i in range(NUM_SMALL_CASES):
            data = self.mnist_70k.sample(n=SMALL_SAMPLE_SIZE).to_numpy()

            # Test agreement with FastPAM1
            _ = bpam_agrees_pam(
                k=SMALL_K_SCHEDULE[i % N_SMALL_K],
                data=data,
                loss="L2",
                test_build=True,
                assert_immediately=True,
                use_fp=True,
            )

            # Test agreement with PAM
            _ = bpam_agrees_pam(
                k=SMALL_K_SCHEDULE[i % N_SMALL_K],
                data=data,
                loss="L2",
                test_build=True,
                assert_immediately=True,
                use_fp=False,
            )

    def test_small_scrna(self):
        """
        Test NUM_SMALL_CASES number of test cases with subsets of size
        SMALL_SAMPLE_SIZE randomly drawn from the full scRNA dataset.
        """
        count = 0
        for i in range(NUM_SMALL_CASES):
            data = self.scrna.sample(n=SMALL_SAMPLE_SIZE).to_numpy()

            # Test agreement with FastPAM1
            count += bpam_agrees_pam(
                k=SMALL_K_SCHEDULE[i % N_SMALL_K],
                data=data,
                loss="L1",
                test_build=True,
                assert_immediately=False,
                use_fp=True,
            )

            # Test agreement with PAM
            count += bpam_agrees_pam(
                k=SMALL_K_SCHEDULE[i % N_SMALL_K],
                data=data,
                loss="L1",
                test_build=True,
                assert_immediately=False,
                use_fp=False,
            )
        # Occasionally some may fail due to degeneracy in the scRNA dataset
        self.assertTrue(count >= 2 * PROPORTION_PASSING * NUM_SMALL_CASES)

    def test_small_mnist_known_cases(self):
        """
        Test BanditPAM on a subset of MNIST with known solutions for both k = 5
        and k = 10, after both the BUILD and SWAP steps.
        """
        kmed_5 = KMedoids(
            n_medoids=5,
            algorithm="BanditPAM",
        )
        kmed_5.seed = 0
        kmed_5.fit(self.small_mnist, "L2")

        self.assertEqual(
            sorted(kmed_5.build_medoids.tolist()), [16, 24, 32, 70, 87]
        )
        self.assertEqual(sorted(kmed_5.medoids.tolist()), [23, 30, 49, 70, 99])

        kmed_10 = KMedoids(
            n_medoids=10,
            algorithm="BanditPAM",
        )
        kmed_10.seed = 0
        kmed_10.fit(self.small_mnist, "L2")

        self.assertEqual(
            sorted(kmed_10.build_medoids.tolist()),
            [16, 24, 32, 49, 70, 82, 87, 90, 94, 99],
        )
        self.assertEqual(
            sorted(kmed_10.medoids.tolist()),
            [16, 25, 31, 49, 63, 70, 82, 90, 94, 99],
        )

    def test_edge_cases(self):
        """
        Test that BanditPAM raises errors on n_medoids being unspecified or an
        empty dataset.
        """
        kmed = KMedoids()
        kmed.seed = 0

        # initialized to empty
        self.assertEqual([], kmed.build_medoids[0].tolist())
        self.assertEqual([], kmed.medoids[0].tolist())

        # error that no value of k has been passed
        self.assertRaises(ValueError, kmed.fit, np.array([]), "L2")

        # error on trying to fit on empy
        self.assertRaises(ValueError, kmed.fit, np.array([]), "L2")


if __name__ == "__main__":
    np.random.seed(0)
    unittest.main()
